;-------------------------------------------------------------------------------
;
;                   PIC 16F84/16CR84
;
;      18 pin Flash/EEPROM 8-bit Microcontroller
;             - 1 K (words) Flash (ROM) program memory
;             - 68 bytes Data RAM
;             - 64 bytes Data EEPROM
;             - 10 Mhz max. Frequency
;             - 35 single word instruction set
;
;-------------------------------------------------------------------------------
;
;                    24C16/24LC16
;
;      16K bit Electrically Erasable PROM
;             - 8 blocks of 256 x 8-bit memory
;             - 2-wire serial interface
;             - a page-write capability for up to 16 bytes of data
;             - 100 kHz max. clock frequency
;
;  Eeprom I/O:
;
;  S = Start Bit (a HIGH to LOW transition of the SDA line while the clock is HIGH)
;  1 = hight level logic
;  0 = low level logic
;  B = Block Address
;  D = Data
;  R = 1Read 0Write
;  A = ACK (0 level)
;  N = NO ACK (1 level)
;  P = Stop Bit (a LOW to HIGH transition of the SDA line while the clock is HIGH)
;
;  BYTE WRITE
;
;  ---Control byte---- ---Address byte-- ---Data byte-----
;  |                 | |               | |               |
;  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2
;  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9
;  | | | | | | | | | | | | | | | | | | | | | | | | | | | | |
;  S 1 0 1 0 B B B R A B B B B B B B B A D D D D D D D D A P
;
;  PAGE WRITE
;
;  ---Control byte---- ---Address byte-- ---Data byte----- ---Data byte-----
;  |                 | |               | |               | |               |
;  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 --- n n n
;  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 --- 1 2 3
;  | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | --- | | |
;  S 1 0 1 0 B B B R A B B B B B B B B A D D D D D D D D A D D D D D --- D A P
;                                        0 0 0 0 0 0 0 0   1 1 1 1 1     k   
;  CURRENT ADDRESS READ
;
;  ---Control byte---- ---Data byte-----
;  |                 | |               |
;  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2
;  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0
;  | | | | | | | | | | | | | | | | | | | |
;  S 1 0 1 0 B B B R A D D D D D D D D N P
;
;  RANDOM READ
;
;  ---Control byte---- ---Address byte-- ---Control byte---- ---Data byte-----
;  |                 | |               | |                 | |               |
;  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3 3 3
;  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9
;  | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | |
;  S 1 0 1 0 B B B R A B B B B B B B B A S 1 0 1 0 B B B R A D D D D D D D D N P
;
;  SEQUENTIAL READ
;
;  ---Control byte---- ---Data byte----- ---Data byte----- ---Data byte-----
;  |                 | |               | |               | |               |
;  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 ----- n n n
;  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 ----- 1 2 3
;  | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | ----- | | |
;  S 1 0 1 0 B B B R A D D D D D D D D A D D D D D D D D A D D D D D ----- D N P
;                      0 0 0 0 0 0 0 0   1 1 1 1 1 1 1 1   2 2 2 2 2       k
;
;-------------------------------------------------------------------------------
;                                                                              *
;       External EEPROM data                                                   *
;                                                                              *
;-------------------------------------------------------------------------------
;
; 0000 Smart Serial number
; byte 0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x90,0x00
;                         |------serial-------|
;
;-------------------------------------------------------------------------------
;


;***********************************************
;define	register function
W		equ	0	;W register Destination
F		equ	1	;FXX register Destination
C		equ	0	;Carry/borrow bit 
DC		equ	1	;Digit carry
Z		equ	2	;Zero flag
PD		equ	3	;Power down
T0		equ	4	;Time-out bit
RP0		equ	5	;Register page select bits
RP1		equ	6	;Not used in 16F8X
IRP		equ	7	;Not used in 16F8X
;
;Bank 0
INDF 	equ	00	;Indirect Address
RTCC		equ	01	;Real Timer Clock Counter
PCL		equ	02	;Program Counter low byte
STATUS		equ	03	;Status	Register
;	d0 = Carry bit
;	d1 = Digit Carry bit
;	d2 = Zero bit
;	d3 = Power Down	bit
;	d4 = Time-Out bit
;		d4	d3	Condition
;		0	0	WDT wake-up from Sleep
;		0	1	WDT time-out (not during Sleep)
;		1	0	MCLR wake-up from Sleep
;		1	1	Power-up
;	d5 = Bank Select 0
;	d6 = n.u
;	d7 = n.u
FSR		equ	04	;File Select Register
;	d0 = select file register 0
;	d1 = select file register 1
;	d2 = select file register 2
;	d3 = select file register 3
;	d4 = select file register 4
;	d5 = 
;	d6 = 
;	d7 =
PORTA		equ	05	;Port A
;	d0 = n.u.(output)
;	d1 = n.u.(output)
;	d2 = n.u.(output)
;	d3 = n.u.(output)
;	d4 = n.u.(output)
;	d5 = n.c.
;	d6 = n.c.
;	d7 = n.c.
PORTB		equ	06	;Port B
;	d0 = n.u.(output)
;	d1 = n.u.(output)
;	d2 = n.u.(output)
;	d3 = n.u.(output)
;	d4 = External EEprom Serial Data
;	d5 = External EEprom Serial Clock
;	d6 = Data Clock	(input)
;	d7 = Data In/Output CAM
EEDATA		equ	08	;Eeprom	data
EECON1		equ	08
EEADR		equ	09	;Eeprom	address
PCLATH		equ	0a	;Program counter high byte
INTCON		equ	0b	;Interrupt control
;	d0 = RB	Port Control
;	d1 = RB0 Interrupt Flag
;	d2 = TMR0 owerflow Flag
;	d3 = RB	Port Change
;	d4 = RB0 Interrupt Enable
;	d5 = TMR0 Overflow Interrupt
;	d6 = EE	Write Complete Interrupt
;	d7 = Global Interrupt Enable
;
;Bank 1
;inad		equ	80	;Indirect Address
;option		equ	81	;Option	Register
;	d0 = Prescaler 0
;	d1 = Prescaler 1
;	d2 = Prescaler 2
;		d3	d2	d1	rtcc Rate	WDT Rate
;		0	0	0	2		1
;		0	0	1	4		2
;		0	1	0	8		4
;		0	1	1	16		8
;		1	0	0	32		16
;		1	0	1	64		32
;		1	1	0	128		64
;		1	1	1	256		128
;	d3 = Prescaler	0=rtcc	1=WDT
;	d4 = rtcc Signal	0=low>high	1=high>low
;	d5 = rtcc source	0=Clkout	1=rtcc pin
;	d6 = Interupt Edge	0=low>high	1=high>low
;	d7 = Pull Up portb	0=Enable	1=Disable
;pcl		equ	82	;Program Counter low byte
;status		equ	83	;Status	Register
;fsr		equ	84	;File Select Register
trisa		equ	85	;Port A	data direction register
trisb		equ	86	;Port B	data direction register
eecon1		equ	88	;Eeprom	Configure Register 1
;	d0 = read
;	d1 = write
;	d2 = write enable
;	d3 = write error
;	d4 = write wait
;	d5 = n.u.
;	d6 = n.u.
;	d7 = n.u.
eecon2		equ	89h	;Eeprom	Configure Register 2
;pclath		equ	8ah	;Program counter high byte
;intcon		equ	8bh	;Interrupt control


            LIST      P=16F84, F=INHX8M
;            include "P16FXX.inc"
            ORG     0x0000
;********************************************************************************
; Vector de reset
;********************************************************************************
            CALL    Delay_A
            TRIS    PORTA	    ; Deshabilita el PORTA
            ADDLW   0x7F	    ; Opciones: Preescaler 1:256 Asigned to WDT
            OPTION		    ; PBPU: PROTB 0 Pullup enable
;********************************************************************************
; Vector de interrupcion
;********************************************************************************

;********************************************************************************
; Secuencia de ATR
;96 40 01 00 11 F7 3B
; 3B F7 11 00 01 40 96 54 30 04 0E 6C B6 D6 90 00
;********************************************************************************
            MOVLW   0x0E	    ; Carga W con 0x0E
            MOVWF   0x10	    ; Registro 0x10 <- 0x0E

Label_0004  MOVF    0x10      , W   ; W <- Registro 0x10
            CALL    Leer_EPI        ; Leer dato 0x0E del EPI
            CALL    SendByte  	    ; Enviar a la CAM
            DECFSZ  0x10      , f   ; Los 16 primeros datos (ATR)
            GOTO    Label_0004	    ; de la RAM

;********************************************************************************
; Send90CAM
;     Envia 90 00 o 90 1D en caso de error
;********************************************************************************
Send90CAM   MOVLW   0x90	    ; Envia 90 a la CAM
            CALL    SendByte  	    ;
Label_0009  MOVLW   0x00	    ;
            BTFSC   PCLATH    , 03
            MOVLW   0x1D
            CALL    SendByte  

;********************************************************************************
; LeeCmd
;    Despues de enviar el OK a la cam espera un nuevo comando
;********************************************************************************
            CALL    ReadByte	    ; C1  (no lo guarda)
            MOVLW   0x0C	    ;
            MOVWF   FSR		    ; R0x0C .. R0x0F : 34 00 00 03
Label_0006  CALL    ReadByte  
            MOVWF   INDF
            INCF    FSR       , f
            BTFSS   FSR       , 04
            GOTO    Label_0006
;********************************************************************************
; ProcCmd
;     Despues de almacenar el comando en 0x0C .. 0x0F lo procesa
;     en 0x0C el comando
;     en 0x0F el nmero de bytes de datos
;********************************************************************************
            MOVLW   0x0D	    ;
            MOVWF   0x10	    ; Inicializa contador a 0x0D
Label_0008  MOVF    0x10      , W   ; Lee eepron interna   d  c  b  a  9  8  7  6  5  4  3  2  1
            ADDLW   0x0E	    ; desde 1B hasta 0E : 5A 1A 40 3C 3A 21 30 20 32 34 12 16 0E 
            CALL    Leer_EPI	    ;                     
            XORWF   0x0C      , W   ; 6E
            BTFSC   STATUS    , Z
            GOTO    Label_0007	    ; Si a encontrado el cmd procesalo
            DECFSZ  0x10      , f   ; decrementa el indice
            GOTO    Label_0008	    ; es comando no es
            MOVLW   0x6D	    ; No reconoce el cmd 
            CALL    SendByte	    ; Lo notifica.
            GOTO    Label_0009
Label_0007  MOVF    0x0C      , W   ; Recupera el comando
            CALL    SendByte	    ; y lo envia a la CAM
            MOVF    0x10      , W   ; Recupera el ndice
            ADDLW   $+2		    ;0x2B le suma el inicio de la tabla
            CLRF    PCLATH	    ;
            MOVWF   PCL		    ;
            GOTO    Comando0E	    ; 0E
            GOTO    Comando16 	    ; 16
            GOTO    Comando12 	    ; 12
            GOTO    Comando34	    ; 34
            GOTO    Comando32	    ; 32
            GOTO    Comando20 	    ; 20
            GOTO    Comando30 	    ; 30
            GOTO    Comando21 	    ; 21
            GOTO    Comando3A 	    ; 3A
            GOTO    Comando3C 	    ; 3C
            GOTO    Comando40 	    ; 40
            GOTO    Comando1A 	    ; 1A
            BSF     PCLATH    , 03  ; 5A


;**************************************************************************************
; Comando0E C1 0E
;   nmero de serie de la tarjeta
;   0E 00 25 00 00 00 00 00 00 90 00
;   Almacenado en la posicin 0 de la eeprom externa
;**************************************************************************************
Comando0E    MOVLW   0x00

;**************************************************************************************
; SendDatEp
;   Lee datos [0x0F] de la eeprom[W] externa y los envia a la CAM
;***************************************************************************************
SendDatEp   CALL    LeeEpToCam
            GOTO    Send90CAM


;***************************************************************************************
Comando1A   INCF    PCLATH    , f
            MOVLW   0xA6
            GOTO    SendDatEp

;***************************************************************************************
; Comando16 C1 16
;    Entidades soportadas
;      16 00 00 07 FF 00 00 FF 90 00
;***************************************************************************************
Comando16   MOVLW   0x08
            GOTO    SendDatEp

;***************************************************************************************
; Comando12 C1 12
;    Identificacin de las entidades soportadas
; 12 00 ee aa aa aa aa aa aa aa aa aa aa aa aa aa aa aa aa 00 xx xx yy dd dd 90 00
; ee   : proveedor
; aa   : ascci nombre del proveedor
; xx xx: direecin de la tarjeta (Smart card addres)
; yy   : desconocido, los cambios no afectan
; dd dd: fecha de la subcripcin bits 0..4 dia, 5..8 mes, 9..15 ao-1990
;***************************************************************************************
Comando12   MOVLW   0xF8
            INCF    0x0D      , f
Label_0019  ADDLW   0x18	    ; Posicin (proveedor)*18 + 10
            BTFSC   STATUS    , Z
            INCF    PCLATH    , f   ; Direccin PCLATH*256 + W
            DECFSZ  0x0D      , f
            GOTO    Label_0019
            GOTO    SendDatEp
;**************************************************************************************
; Comando3A C1 3A 00 00 10
;    Enviar la palabra de control
;    La palabra de control desncriptada se encuentra en el registro 0x30 y sucesivos
;**************************************************************************************
Comando3A   MOVLW   0x30
            MOVWF   FSR
Label_001A  MOVF    INDF      , W
            CALL    SendByte  
            INCF    FSR       , f
            DECFSZ  0x0F      , f
            GOTO    Label_001A
            GOTO    Send90CAM

;**************************************************************************************
; Comando3C C1 3C P1 P2 LL
;
;**************************************************************************************
Comando3C   CALL    Label_001B
            MOVF    0x14      , W
            SUBLW   0x1D
            MOVWF   EEADR
            MOVLW   0x0C
            CALL    Label_001C
            CALL    Label_001D
            MOVF    0x14      , W
            SUBLW   0x25
            MOVWF   EEADR
            MOVLW   0x0C
            CALL    Label_001C
            CALL    Label_001E
            CALL    Label_001F
            CALL    Label_0020
            CALL    Label_001F
            CALL    Label_0020
            GOTO    Send90CAM

Label_001B  MOVF    0x0F      , W   ; Nr de bytes que vamos a recibir
            MOVWF   0x10	    ; R0x10 ser el contador
Label_0023  MOVLW   0x31	    ; 49 datos
            SUBWF   0x10      , W   ; Recibidos - 49
            BTFSS   STATUS    , C   ; Si resultado positivo skip
            GOTO    Label_0021	    ; Si hay mas de 0x30 datos
            MOVLW   0x08	    ; ves despreciandolos de 8 en 8
            SUBWF   0x10      , f   ; hasta quedarte con menos de 0x30
            MOVWF   0x0F
            CALL    RecNBytes
            GOTO    Label_0023

Label_0021  MOVLW   0x18
            MOVWF   FSR		    ; almacena el nr de bytes indicado en 0x10
Label_0024  CALL    ReadByte	    ; en el registro 0x18
            MOVWF   INDF
            INCF    FSR       , f
            DECFSZ  0x10      , f
            GOTO    Label_0024

            MOVF    FSR       , W   ; Almacena ultima pos + 1 en R0x14
            MOVWF   0x14
            CALL    CalcPtr         ; Calcula la direccion de la Key en la eeprom [PCLATH]*256 + [0x10]
            MOVLW   0x40	    ;
            MOVWF   FSR
            CALL    LeeBckEpe	    ; Lee la key y la almacena en R0x04
            MOVLW   0x40	    ;
            BTFSC   0x0D      , 03  ; 
            ADDWF   0x10      , f   ;
            CALL    LeeBckEpe	    ;
            BTFSS   PCLATH    , 04
            GOTO    Label_0027
            MOVLW   0x18
            CALL    Label_0028
            CALL    Label_001F
            MOVLW   0x18
            CALL    Label_0028
            MOVLW   0x10
            CALL    Label_0028
            CALL    Label_001F
            MOVLW   0x10
            CALL    Label_0028
            MOVLW   0x08
            CALL    Label_0028
            CALL    Label_001F
            MOVLW   0x08
            CALL    Label_0028
            CALL    Label_0029
Label_0029  CALL    Label_0020
            CALL    Label_001F
Label_0027  RETURN
Label_0020  MOVLW   0xF8
Label_0028  MOVWF   EEADR
            MOVLW   0x30
            CALL    Label_001C
            RETURN
Label_001F  CLRF    0x0F
            MOVLW   0x02
            MOVWF   PCLATH
            CLRF    0x0D
Label_002B  CALL    Label_002A
            INCF    0x0F      , f
            MOVF    0x0F      , W
            ANDLW   0x03
            BTFSC   STATUS    , Z
            INCF    0x0D      , f
            BTFSS   0x0F      , 06
            GOTO    Label_002B
            MOVLW   0x0F
            MOVWF   0x0D
            MOVLW   0x0C
            MOVWF   0x0F
Label_0030  CLRW
            CALL    Label_002C
            XORWF   0x34      , W
            CALL    GetTabByt    
            MOVWF   0x10
            MOVLW   0x01
            CALL    Label_002C
            XORWF   0x35      , W
            CALL    GetTabByt    
            MOVWF   0x11
            MOVLW   0x02
            CALL    Label_002C
            XORWF   0x36      , W
            CALL    GetTabByt    
            MOVWF   0x12
            MOVLW   0x03
            CALL    Label_002C
            XORWF   0x37      , W
            CALL    GetTabByt    
            MOVWF   0x13
            INCF    PCLATH    , f
            XORWF   0x10      , f
            SWAPF   0x13      , W
            ADDWF   0x12      , W
            CALL    LeeRandom
            MOVWF   0x12
            XORWF   0x13      , f
            SWAPF   0x12      , W
            ADDWF   0x11      , W
            CALL    LeeRandom
            MOVWF   0x11
            XORWF   0x12      , f
            SWAPF   0x11      , W
            ADDWF   0x10      , W
            CALL    LeeRandom
            MOVWF   0x10
            DECF    PCLATH    , f
            XORWF   0x11      , f
            SWAPF   0x10      , W
            ADDWF   0x13      , W
            CALL    GetTabByt    
            MOVWF   0x13
            XORWF   0x10      , f
            SWAPF   0x13      , W
            ADDWF   0x12      , W
            CALL    GetTabByt    
            MOVWF   0x12
            XORWF   0x13      , f
            SWAPF   0x12      , W
            ADDWF   0x11      , W
            CALL    GetTabByt    
            MOVWF   0x11
            XORWF   0x12      , f
            MOVF    0x10      , W
            XORWF   0x11      , f
            INCF    PCLATH    , f
            MOVF    0x11      , W
            CALL    LeeRandom
            XORWF   0x30      , f
            MOVF    0x13      , W
            CALL    LeeRandom
            XORWF   0x31      , f
            MOVF    0x10      , W
            CALL    LeeRandom
            XORWF   0x32      , f
            MOVF    0x12      , W
            CALL    LeeRandom
            XORWF   0x33      , f
            DECF    PCLATH    , f
            MOVLW   0x03
            ADDWF   0x0F      , f
            CALL    Label_002A
            DECF    0x0F      , f
            CALL    Label_002A
            DECF    0x0F      , f
            CALL    Label_002A
            DECF    0x0F      , f
            CALL    Label_002A
            MOVF    0x0D      , f
            BTFSC   STATUS    , Z
            RETURN
            MOVLW   0x04
            MOVWF   EEADR
            MOVLW   0x34
            CALL    Label_002F
            DECF    0x0D      , f
            MOVLW   0xFC
            ADDWF   0x0F      , f
            MOVLW   0x0C
            BTFSC   0x0F      , 07
            MOVWF   0x0F
            GOTO    Label_0030
Label_002A  MOVLW   0x01
            CALL    Label_002C
            MOVWF   0x0C
            MOVLW   0xFF
            CALL    Label_002C
            XORWF   0x0C      , f
            CLRW
            CALL    Label_002C
            MOVF    0x0C      , W
            XORWF   0x0D      , W
            CALL    GetTabByt    
            XORWF   INDF      , f
            RETURN

Label_002C  ADDWF   0x0F      , W
            ANDLW   0x0F
            MOVWF   FSR
            BSF     FSR       , 06
            MOVF    INDF      , W
            RETURN

CalcPtr     CLRF    0x10	    ; Borra registro 0x10
            RRF     0x0D      , f   ; Proveedor
            RRF     0x10      , f
            RRF     0x0D      , W
            RRF     0x10      , f
            ANDLW   0x03
            MOVWF   PCLATH	    ; W = x3 x2 del proveedor
            BSF     PCLATH    , 02  ; PCLATH = 1 x3 x2
            BTFSC   0x0E      , 07  ; Si Key < 7F skip
            BSF     PCLATH    , 04  ; si es mayor de 7F PCLATCH = 11 x3 x2
            MOVLW   0x07	    ;
            ANDWF   0x0E      , f   ; Nos quedamos con el nible bajo de key 0..f
Label_0031  MOVLW   0x08	    ;
            DECF    0x0E      , f   
            BTFSC   0x0E      , 07  ;
            RETURN
            ADDWF   0x10      , f   ;
            GOTO    Label_0031	    ;

;****************************************************************************
;  LeeBckEpe
;   Lee un bloque de 8 bytes de la eeprom externa. La inicializa con un ciclo 
;   de y despues lee un bloque de datos.
;   PCLATH: pgina
;   R 0x10: Direccin
;   Llamando a LeeBckEpeW W: Direccin
;   Salida: Los datos se almacenan el los registros apuntados por FSR
;****************************************************************************
LeeBckEpe   MOVF    0x10      , W
LeeBckEpeW  CALL    WriteInit   
            MOVLW   0x08
            MOVWF   0x11
ReadLoop    CALL    LeeEpe    
            MOVWF   INDF
            INCF    FSR       , f
            DECFSZ  0x11      , f
            GOTO    ReadLoop    
            RETURN

;****************************************************************************
; Comando34
;     Respuesta al comando 34
;****************************************************************************
Comando34   CALL    RecNBytes
            GOTO    Send90CAM

Comando32   INCF    PCLATH    , f
            MOVLW   0x99
            BTFSS   0x0F      , 00
            GOTO    Label_0035
            GOTO    SendDatEp
Comando20   MOVF    0x0F      , W
            MOVLW   0x38
            MOVWF   FSR
Label_0036  CALL    ReadByte  
            MOVWF   INDF
            INCF    FSR       , f
            BTFSS   FSR       , 06
            GOTO    Label_0036
            MOVF    0x0D      , W
            ADDWF   PCLATH    , f
            MOVF    0x0E      , W
            CALL    WriteInit   
            CALL    Label_0037
            GOTO    Send90CAM
Label_0037  MOVLW   0x37
            MOVWF   FSR
Label_0039  INCF    FSR       , f
            MOVF    INDF      , W
            CALL    WriteEpeA
            DECFSZ  0x0F      , f
            GOTO    Label_0039
            GOTO    StopBit    

;*******************************************************************************
;  Rec0F_W
;    Resta al nr de bytes recibidos W y desprecia todos menos los W ltimos
;*******************************************************************************
Rec0F_W     SUBWF   0x0F      , f

;*******************************************************************************
; RecNBytes
; Ignora los tres bytes recibidos de la instruccion 34
;*******************************************************************************
RecNBytes   CALL    ReadByte  
            DECFSZ  0x0F      , f
            GOTO    RecNBytes
            RETURN

;********************************************************************************
; SendByte 
;    envia un byte y espera uno de respuesta
;    Entrada W Byte a enviar
;    Salida  W Byet recibido
;********************************************************************************
SendByte    MOVWF   0x16	    ; Gurada W en 0x16	
            MOVLW   0x32	    ; Retardo de 0x32
            CALL    Delay_W
            MOVLW   0x7F	    ; 0111111
            TRIS    PORTB	    ; Escritura de la CAM
            CLRF    0x17	    ; Borrar registro 0x17
            BCF     PORTB     , 07  ; Envia 0 a la CAM
            MOVLW   0x08	    ; Se van a escribir 8 bits
            MOVWF   0x14	    ; El contador al registro 0x14
Label_003C  CALL    Delay_A	    
            MOVF    0x16      , W   ; Recupera el contenido de W
            XORWF   0x17      , f   ; W XOR R0x17 -> Rx017
            BCF     PORTB     , 07  ; Envia 0 a la CAM
            BTFSC   0x16      , 00  ; Si bit 0 de W es 0 Saltar a la siguiente
            BSF     PORTB     , 07  ; 
            RRF     0x16      , f
            DECFSZ  0x14      , f
            GOTO    Label_003C
            CALL    Delay_A
            BCF     PORTB     , 07
            BTFSC   0x17      , 00
            BSF     PORTB     , 07
            CALL    Delay_A
            MOVLW   0xFF
            TRIS    PORTB
            MOVLW   0x4F
            GOTO    Delay_W

;*****************************************************************************
; ReadByte Espera hata que se recibe un byte de la cam.
;     Salida W byte leido
;*****************************************************************************      
ReadByte    BTFSC   PORTB     , 07
            GOTO    ReadByte  
            MOVLW   0x2A
            CALL    Delay_W
            MOVLW   0x09
            MOVWF   0x14
Label_003D  BCF     STATUS    , C
            BTFSC   PORTB     , 07
            BSF     STATUS    , C
            RRF     0x16      , f
            CALL    Delay_A
            DECFSZ  0x14      , f
            GOTO    Label_003D
            RLF     0x16      , f
            MOVLW   0x28
            CALL    Delay_W
            MOVF    0x16      , W
            RETURN

;****************************************************************
; WriteCW
;    Escribe el control word en la eeprom externa
;****************************************************************
WriteCW     CALL    StartBit    
            BCF     STATUS    , C
            CALL    EpCtrlWrd 
            BTFSC   STATUS    , C   ;ACK = 0 sino repite Control Word
            GOTO    WriteCW	    
            RETURN

;****************************************************************
; function : Delay												*
;            Sub: Delay_A  time 0x1A     	
;			      Delay_W  time in W     	
;****************************************************************
Delay_A     MOVLW   0x1A
Delay_W     MOVWF   0x15
Delay_L     DECFSZ  0x15      , f
            GOTO    Delay_L
            RETLW   0x00

;****************************************************************
; Leer_EPI 
;   Leer EEPROM interna	
;   W Direccion a leer		      	
;   Salida W Dato		       
;****************************************************************
Leer_EPI    MOVWF   EEADR		; Direccion de EPI <- W
            BSF     STATUS    , RP0 	; Selecciona pagina 1
            BSF     EECON1    , 00	; Se va a escribir la EPI 
            BCF     STATUS    , RP0 	; 
            MOVF    EEDATA    , W	; W <- EPI0[W]
            RETURN

;****************************************************************
; LeeEpToCam
;    Lee la eeprom externa y envia los datos a la CAM
;****************************************************************
LeeEpToCam  CALL    WriteInit   
ETCLoop	    DECF    0x0F      , f
            BTFSC   0x0F      , 07
            RETURN
            CALL    LeeEpe    
            CALL    SendByte  
            GOTO    ETCLoop
Label_001C  CALL    Label_002F
            GOTO    Label_0043
Label_002F  MOVWF   FSR
Label_0043  CALL    Label_0044
            CALL    Label_0044
            CALL    Label_0044
Label_0044  MOVF    INDF      , W
            MOVWF   EEDATA
            MOVF    EEADR     , W
            SUBWF   FSR       , f
            MOVF    EEDATA    , W
            XORWF   INDF      , f
            MOVF    INDF      , W
            XORWF   EEDATA    , f
            MOVF    EEDATA    , W
            XORWF   INDF      , f
            MOVF    EEADR     , W
            ADDWF   FSR       , f
            MOVF    EEDATA    , W
            MOVWF   INDF
            INCF    FSR       , f
            RETURN
;*********************************************************************
;  LeeRandom
;     Lee la eeprom externa iniciando una escritura para cargar la 
;     direccin y despues realiza una lectura de la posicin actual
;     PCLATCH Pgina
;     EADR    Direccin
;     Salida: W dato ledo
;*********************************************************************
LeeRandom   MOVWF   EEADR
            CALL    StartBit    
            BCF     STATUS    , C
            CALL    EpCtrlWrd 
            MOVF    EEADR     , W
            CALL    WriteEpeA
;           call    LeeEpe	    ; No se utiliza por estar a continuacin

;*********************************************************************
;  LeeEpe
;     Realiza el ciclo completo de lectura de la eeprom externa
;     Lee el byte actual de la eeprom. Solo le envia la pgina.
;     Pgina en PCLATH
;     W dato
;*********************************************************************
LeeEpe      CALL    StartBit    
            BSF     STATUS    , C
            CALL    EpCtrlWrd 
            CALL    LeeByteepe
            CALL    StopBit    
            MOVF    EEDATA    , W
            RETURN

;********************************************************************
; WriteInit  
;    Lee un byte de la eeprom externa
;    [PCLATH]*256 + W   direccion.
;********************************************************************    
WriteInit   MOVWF   EEADR	    ; Carga la pgina.
            CALL    WriteCW	    ; Inicia ciclo de escritura
            MOVF    EEADR     , W   ; carga la direccin
            GOTO    WriteEpeA	    ; Envia la direccin 

;********************************************************************
;  DataToEpeF
;     Escribe un byte en la eeprom externa y termina el ciclo de 
;     la eeprom con un stop bit
;********************************************************************
DataToEpeF  CALL    WriteEpeA
            GOTO    StopBit    

;********************************************************************
; EpCtrlWrd
; Palabra de control
;********************************************************************
EpCtrlWrd  RLF     PCLATH    , W   ; Carga el selector de pagina
            ANDLW   0x0F	    ; rotado un bit y se queda 4lsb
            IORLW   0xA0	    ;

WriteEpeA   MOVWF   EEDATA	    ; 
            MOVLW   0xCF	    ; W EEPE
            TRIS    PORTB
            CALL    Escrib4bits
            CALL    Escrib4bits
            GOTO    AckFromEpe	    ; y return

;*********************************************************************
;  Escrib4bits
;      Envia a la eprom 4 bits rotando EEDATA
;*********************************************************************
Escrib4bits CALL    Escrib1bit 
            CALL    Escrib1bit 
            CALL    Escrib1bit 
Escrib1bit  RLF     EEDATA    , f
            GOTO    SendCarry	    ; y return

;********************************************************************
;  LeeByteepe
;     Obtiene un byte de datos de la eeprom externa
;     una vez inicializada
;********************************************************************
LeeByteepe  MOVLW   0xDF
            TRIS    PORTB
            CALL    Lee4bits  
            CALL    Lee4bits  
            BSF     STATUS    , C
            GOTO    SendCarryI    
Lee4bits    CALL    Lee1bit     
            CALL    Lee1bit     
            CALL    Lee1bit     
Lee1bit     CALL    AckFromEpei
            RLF     EEDATA    , f
            RETURN

;*********************************************************
; StartBit
;    Envia un 1 a la eeprom externa.
;*********************************************************
StartBit    BSF     PORTB     , 04  ; SData 1
            MOVLW   0xCF
            TRIS    PORTB
            BSF     PORTB     , 05  ; SCLK 1 
            NOP
            BCF     PORTB     , 04  ; Sdata 0
            GOTO    SCLK     
;*********************************************************
; StopBit
;    Envia un cero a la eeprom externa
;*********************************************************
StopBit     BCF     PORTB     , 04  ; Sdata 0
            MOVLW   0xCF	    ; CAMR CLKR SCLKW SDataW
            TRIS    PORTB
            NOP
            BSF     PORTB     , 05  ; SCLK <- 1
            NOP
            BSF     PORTB     , 04  ; 

SCLKnop     NOP

SCLK        BCF     PORTB     , 05
            RETURN

;*********************************************************
; SendCarryI
;    Envia a la eeprom externa el bit de acarreo
;*********************************************************
SendCarryI  MOVLW   0xCF
            TRIS    PORTB
SendCarry   BTFSS   STATUS    , C
            BCF     PORTB     , 04
            BTFSC   STATUS    , C
            BSF     PORTB     , 04
            BSF     PORTB     , 05
            GOTO    SCLKnop
;*********************************************************
; AckFromEpe
;     Lee un bit de la eeprom externa y lo almacena en C
;*********************************************************
AckFromEpe    MOVLW   0xDF
            TRIS    PORTB
AckFromEpei BSF     PORTB     , 05
            NOP
            BSF     STATUS    , C
            BTFSS   PORTB     , 04
            BCF     STATUS    , C
            GOTO    SCLK     

;**********************************************************
;  Tabla para el algoritmo de desencriptar
;**********************************************************
GetTabByt   MOVWF   PCL
            RETLW   0x2A
            RETLW   0xE1
            RETLW   0x0B
            RETLW   0x13
            RETLW   0x3E
            RETLW   0x6E
            RETLW   0x32
            RETLW   0x48
            RETLW   0xD3
            RETLW   0x31
            RETLW   0x08
            RETLW   0x8C
            RETLW   0x8F
            RETLW   0x95
            RETLW   0xBD
            RETLW   0xD0
            RETLW   0xE4
            RETLW   0x6D
            RETLW   0x50
            RETLW   0x81
            RETLW   0x20
            RETLW   0x30
            RETLW   0xBB
            RETLW   0x75
            RETLW   0xF5
            RETLW   0xD4
            RETLW   0x7C
            RETLW   0x87
            RETLW   0x2C
            RETLW   0x4E
            RETLW   0xE8
            RETLW   0xF4
            RETLW   0xBE
            RETLW   0x24
            RETLW   0x9E
            RETLW   0x4D
            RETLW   0x80
            RETLW   0x37
            RETLW   0xD2
            RETLW   0x5F
            RETLW   0xDB
            RETLW   0x04
            RETLW   0x7A
            RETLW   0x3F
            RETLW   0x14
            RETLW   0x72
            RETLW   0x67
            RETLW   0x2D
            RETLW   0xCD
            RETLW   0x15
            RETLW   0xA6
            RETLW   0x4C
            RETLW   0x2E
            RETLW   0x3B
            RETLW   0x0C
            RETLW   0x41
            RETLW   0x62
            RETLW   0xFA
            RETLW   0xEE
            RETLW   0x83
            RETLW   0x1E
            RETLW   0xA2
            RETLW   0x01
            RETLW   0x0E
            RETLW   0x7F
            RETLW   0x59
            RETLW   0xC9
            RETLW   0xB9
            RETLW   0xC4
            RETLW   0x9D
            RETLW   0x9B
            RETLW   0x1B
            RETLW   0x9C
            RETLW   0xCA
            RETLW   0xAF
            RETLW   0x3C
            RETLW   0x73
            RETLW   0x1A
            RETLW   0x65
            RETLW   0xB1
            RETLW   0x76
            RETLW   0x84
            RETLW   0x39
            RETLW   0x98
            RETLW   0xE9
            RETLW   0x53
            RETLW   0x94
            RETLW   0xBA
            RETLW   0x1D
            RETLW   0x29
            RETLW   0xCF
            RETLW   0xB4
            RETLW   0x0D
            RETLW   0x05
            RETLW   0x7D
            RETLW   0xD1
            RETLW   0xD7
            RETLW   0x0A
            RETLW   0xA0
            RETLW   0x5C
            RETLW   0x91
            RETLW   0x71
            RETLW   0x92
            RETLW   0x88
            RETLW   0xAB
            RETLW   0x93
            RETLW   0x11
            RETLW   0x8A
            RETLW   0xD6
            RETLW   0x5A
            RETLW   0x77
            RETLW   0xB5
            RETLW   0xC3
            RETLW   0x19
            RETLW   0xC1
            RETLW   0xC7
            RETLW   0x8E
            RETLW   0xF9
            RETLW   0xEC
            RETLW   0x35
            RETLW   0x4B
            RETLW   0xCC
            RETLW   0xD9
            RETLW   0x4A
            RETLW   0x18
            RETLW   0x23
            RETLW   0x9F
            RETLW   0x52
            RETLW   0xDD
            RETLW   0xE3
            RETLW   0xAD
            RETLW   0x7B
            RETLW   0x47
            RETLW   0x97
            RETLW   0x60
            RETLW   0x10
            RETLW   0x43
            RETLW   0xEF
            RETLW   0x07
            RETLW   0xA5
            RETLW   0x49
            RETLW   0xC6
            RETLW   0xB3
            RETLW   0x55
            RETLW   0x28
            RETLW   0x51
            RETLW   0x5D
            RETLW   0x64
            RETLW   0x66
            RETLW   0xFC
            RETLW   0x44
            RETLW   0x42
            RETLW   0xBC
            RETLW   0x26
            RETLW   0x09
            RETLW   0x74
            RETLW   0x6F
            RETLW   0xF7
            RETLW   0x6B
            RETLW   0x4F
            RETLW   0x2F
            RETLW   0xF0
            RETLW   0xEA
            RETLW   0xB8
            RETLW   0xAE
            RETLW   0xF3
            RETLW   0x63
            RETLW   0x6A
            RETLW   0x56
            RETLW   0xB2
            RETLW   0x02
            RETLW   0xD8
            RETLW   0x34
            RETLW   0xA4
            RETLW   0x00
            RETLW   0xE6
            RETLW   0x58
            RETLW   0xEB
            RETLW   0xA3
            RETLW   0x82
            RETLW   0x85
            RETLW   0x45
            RETLW   0xE0
            RETLW   0x89
            RETLW   0x7E
            RETLW   0xFD
            RETLW   0xF2
            RETLW   0x3A
            RETLW   0x36
            RETLW   0x57
            RETLW   0xFF
            RETLW   0x06
            RETLW   0x69
            RETLW   0x54
            RETLW   0x79
            RETLW   0x9A
            RETLW   0xB6
            RETLW   0x6C
            RETLW   0xDC
            RETLW   0x8B
            RETLW   0xA7
            RETLW   0x1F
            RETLW   0x90
            RETLW   0x03
            RETLW   0x17
            RETLW   0x1C
            RETLW   0xED
            RETLW   0xD5
            RETLW   0xAA
            RETLW   0x5E
            RETLW   0xFE
            RETLW   0xDA
            RETLW   0x78
            RETLW   0xB0
            RETLW   0xBF
            RETLW   0x12
            RETLW   0xA8
            RETLW   0x22
            RETLW   0x21
            RETLW   0x3D
            RETLW   0xC2
            RETLW   0xC0
            RETLW   0xB7
            RETLW   0xA9
            RETLW   0xE7
            RETLW   0x33
            RETLW   0xFB
            RETLW   0xF1
            RETLW   0x70
            RETLW   0xE5
            RETLW   0x17
            RETLW   0x96
            RETLW   0xF8
            RETLW   0x8D
            RETLW   0x46
            RETLW   0xA1
            RETLW   0x86
            RETLW   0xE2
            RETLW   0x40
            RETLW   0x38
            RETLW   0xF6
            RETLW   0x68
            RETLW   0x25
            RETLW   0x16
            RETLW   0xAC
            RETLW   0x61
            RETLW   0x27
            RETLW   0xCB
            RETLW   0x5B
            RETLW   0xC8
            RETLW   0x2B
            RETLW   0x0F
            RETLW   0x99
            RETLW   0xDE
            RETLW   0xCE
            RETLW   0xC5
;********************************************************************
;  Comando30
;    Control del PIN
;********************************************************************
Comando30   MOVLW   0x02
            CALL    Rec0F_W	    ; Solo interesan los dos ltimos bytes
            CALL    ReadByte	    ; Recibe el penultimo
            IORWF   0x0F      , f   ; 
            ANDLW   0x0F
            MOVWF   0x0C
            CALL    ReadByte  
            MOVWF   0x0D
            IORWF   0x0F      , f
            MOVF    0x0E      , f
            BTFSS   STATUS    , Z
            GOTO    Send90CAM
            INCF    PCLATH    , f
            MOVLW   0x90
            MOVF    0x0F      , f
            BTFSC   STATUS    , Z
            GOTO    Label_0051
            CALL    LeeRandom
            MOVWF   0x0F
            BTFSS   0x0F      , 06
            GOTO    Label_0052
            BTFSC   0x0F      , 03
            GOTO    Label_0053
            XORLW   0x40
            CALL    LeeRandom
            IORWF   0x0C      , f
            SWAPF   0x0C      , f
            MOVF    0x0F      , W
            XORLW   0x40
            INCF    0x0F      , f
Label_0052  CALL    WriteInit   
            MOVF    0x0D      , W
            CALL    WriteEpeA
            SWAPF   0x0C      , W
            CALL    DataToEpeF
            MOVLW   0x40
            XORWF   0x0F      , W
Label_0051  MOVWF   0x0F
            INCF    0x0F      , f
            MOVLW   0x90
            CALL    WriteInit   
            MOVF    0x0F      , W
            CALL    DataToEpeF
            GOTO    Send90CAM
Label_0053  MOVLW   0x38
            MOVWF   FSR
            MOVLW   0x91
            CALL    LeeBckEpeW
            MOVF    0x0C      , W
            IORWF   0x3F      , f
            CLRF    PCLATH
            MOVF    0x0D      , W
            ANDLW   0x07
            ADDWF   PCLATH    , f
            MOVF    0x0D      , W
            ANDLW   0xF8
            CALL    WriteInit   
            MOVLW   0x08
            MOVWF   0x0F
            CALL    Label_0037
            GOTO    Send90CAM
Comando40   MOVF    0x0D      , W
            MOVWF   0x17
            BSF     PCLATH    , 03
            MOVLW   0x34
            SUBWF   0x0F      , W
            BTFSS   STATUS    , C
            GOTO    Comando34
            CALL    Label_001B
            BSF     PCLATH    , 03
            MOVF    0x17      , W
            ANDLW   0x0F
            BTFSC   STATUS    , Z
            GOTO    Send90CAM
            MOVLW   0x13
Label_0058  SUBWF   0x14      , f
            MOVF    0x14      , W
            MOVWF   FSR
            MOVLW   0x18
            SUBWF   FSR       , W
            BTFSS   STATUS    , C
            GOTO    Label_0056
            MOVF    INDF      , W
            XORLW   0x90
            BTFSS   STATUS    , Z
            GOTO    Send90CAM
            MOVF    FSR       , W
            SUBLW   0x0A
            MOVWF   EEADR
            MOVLW   0x0C
            CALL    Label_001C
            CALL    Label_001E
            CALL    Label_001D
            CALL    Label_001F
            CALL    Label_0020
            CALL    Label_0057
            MOVLW   0x0A
            GOTO    Label_0058
Label_0056  ADDLW   0x1F
            MOVWF   FSR
            MOVF    INDF      , W
            XORLW   0x21
            BTFSS   STATUS    , Z
            GOTO    Send90CAM
            CLRF    PCLATH
            MOVLW   0x0E
            INCF    0x17      , f
Label_0059  ADDLW   0x18
            BTFSC   STATUS    , C
            INCF    PCLATH    , f
            DECFSZ  0x17      , f
            GOTO    Label_0059
            MOVWF   0x0C
            CALL    WriteInit   
            INCF    FSR       , f
            CALL    LeeEpe    
            XORWF   INDF      , W
            BTFSS   STATUS    , Z
            GOTO    Label_005A
            INCF    FSR       , f
            CALL    LeeEpe    
            XORWF   INDF      , W
            BTFSC   STATUS    , Z
            GOTO    Send90CAM
            DECF    FSR       , f
Label_005A  MOVF    0x0C      , W
            CALL    WriteInit   
            MOVF    INDF      , W
            CALL    WriteEpeA
            INCF    FSR       , f
            MOVF    INDF      , W
            CALL    DataToEpeF
            GOTO    Send90CAM
Label_0057  INCF    0x14      , W
            MOVWF   FSR
            MOVF    INDF      , W
            MOVWF   0x0E
            MOVF    0x17      , W
            MOVWF   0x0D
            CALL    CalcPtr     
            MOVF    0x10      , W
            CALL    WriteInit   
            MOVLW   0x38
            MOVWF   FSR
Label_005C  CALL    LeeEpe    
            XORWF   INDF      , W
            BTFSS   STATUS    , Z
            GOTO    Label_005B
            INCF    FSR       , f
            BTFSC   FSR       , 03
            GOTO    Label_005C
            RETURN
Label_005B  MOVF    0x10      , W
            CALL    WriteInit   
            MOVLW   0x08
            MOVWF   0x0F
            CALL    Label_0037
            RETURN
Label_001E  MOVLW   0x24
            MOVWF   EEADR
            MOVLW   0x30
            CALL    Label_001C
            RETURN
Label_001D  MOVLW   0x2C
            MOVWF   EEADR
            MOVLW   0x38
            CALL    Label_001C
            RETURN

Label_0035  MOVLW   0x83
            CALL    SendByte  
            MOVLW   0x0A
            SUBWF   0x0F      , W
            MOVWF   0x10
            MOVLW   0x08
            MOVWF   0x0F
            BCF     STATUS    , C
            RLF     0x0D      , f
            RLF     0x0D      , f
            RLF     0x0D      , W
            INCF    PCLATH    , f
            CALL    LeeEpToCam
            MOVLW   0x04
            CALL    SendByte  
Label_005D  DECF    0x10      , f
            BTFSC   0x10      , 07
            GOTO    Send90CAM
            MOVLW   0xFF
            CALL    SendByte  
            GOTO    Label_005D
Comando21   MOVF    0x0D      , W
            ADDWF   PCLATH    , f
            MOVF    0x0E      , W
            CALL    WriteInit   
Label_005E  CALL    LeeEpe    
            CALL    SendByte  
            DECFSZ  0x0F      , f
            GOTO    Label_005E
            GOTO    Send90CAM
 
 ;           ORG     0x2000
 ;           DATA    0x0F
 ;           DATA    0x0F
 ;           DATA    0x0F
 ;           DATA    0x0F
 ;
 ;           ORG     0x2007
 ;           DATA    0x09

; Datos de la eeprom interna 
            ORG     0x2100
            DATA    0x00
            DATA    0xD6
            DATA    0xB6
            DATA    0x6C
            DATA    0x0E
            DATA    0x04
            DATA    0x30
            DATA    0x54
            DATA    0x96
            DATA    0x40
            DATA    0x01
            DATA    0x00
            DATA    0x11
            DATA    0xF7
            DATA    0x3B
            DATA    0x0E
            DATA    0x16
            DATA    0x12
            DATA    0x34
            DATA    0x32
            DATA    0x20
            DATA    0x30
            DATA    0x21
            DATA    0x3A
            DATA    0x3C
            DATA    0x40
            DATA    0x1A
            DATA    0x5A
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF
            DATA    0xFF

            END
